@@ -1,14 +1,15 @@
 use super::encodings::{self, bytes_to_string, string_to_bytes};
 use super::{Bookmark, Dictionary, Object, ObjectId};
+use crate::encryption;
 use crate::xref::{Xref, XrefType};
 use crate::{Error, Result, Stream};
-use encoding::all::UTF_16BE;
-use encoding::types::{DecoderTrap, EncoderTrap, Encoding};
+use encoding_rs::UTF_16BE;
 use log::info;
 use std::cmp::max;
 use std::collections::{BTreeMap, HashMap};
 use std::io::Write;
 use std::str;
+use crate::xobject::PdfImage;
 
 /// A PDF document.
 ///
@@ -196,6 +197,19 @@ impl Document {
         self.get_object_mut(id).and_then(Object::as_dict_mut)
     }
 
+    /// Get dictionary in dictionary by key.
+    pub fn get_dict_in_dict<'a>(&'a self, node: &'a Dictionary, key: &[u8]) -> Result<&'a Dictionary> {
+        match node.get(key)? {
+            Object::Reference(object_id) => {
+                self.get_dictionary(*object_id)
+            }
+            Object::Dictionary(dic) => {
+                Ok(dic)
+            }
+            _ => { Err(Error::Type) }
+        }
+    }
+
     /// Traverse objects from trailer recursively, return all referenced object IDs.
     pub fn traverse_objects<A: Fn(&mut Object)>(&mut self, action: A) -> Vec<ObjectId> {
         fn traverse_array<A: Fn(&mut Object)>(array: &mut [Object], action: &A, refs: &mut Vec<ObjectId>) {
@@ -234,6 +248,77 @@ impl Document {
         refs
     }
 
+    /// Return dictionary with encryption information
+    pub fn get_encrypted(&self) -> Result<&Dictionary> {
+        self.trailer
+            .get(b"Encrypt")
+            .and_then(Object::as_reference)
+            .and_then(|id| self.get_dictionary(id))
+    }
+
+    /// Return true is PDF document is encrypted
+    pub fn is_encrypted(&self) -> bool {
+        self.get_encrypted().is_ok()
+    }
+
+    /// Replaces all encrypted Strings and Streams with their decrypted contents
+    pub fn decrypt<P: AsRef<[u8]>>(&mut self, password: P) -> Result<()> {
+        // Find the ID of the encryption dict; we'll want to skip it when decrypting
+        let encryption_obj_id = self.trailer.get(b"Encrypt").and_then(Object::as_reference)?;
+
+        // Since PDF 1.5, metadata may or may not be encrypted; defaults to true
+        let metadata_is_encrypted = self
+            .get_object(encryption_obj_id)?
+            .as_dict()?
+            .get(b"EncryptMetadata")
+            .and_then(|o| o.as_bool())
+            .unwrap_or(true);
+
+        let key = encryption::get_encryption_key(self, &password, true)?;
+        for (&id, obj) in self.objects.iter_mut() {
+            // The encryption dictionary is not encrypted, leave it alone
+            if id == encryption_obj_id {
+                continue;
+            }
+
+            // If a Metadata stream but metadata isn't encrypted, leave it alone
+            if obj.type_name().unwrap_or("") == "Metadata" && !metadata_is_encrypted {
+                continue;
+            }
+
+            let decrypted = match encryption::decrypt_object(&key, id, &*obj) {
+                Ok(content) => content,
+                Err(encryption::DecryptionError::NotDecryptable) => {
+                    continue;
+                }
+                Err(_err) => {
+                    return Err(_err.into());
+                }
+            };
+
+            // Only strings and streams are encrypted
+            match obj {
+                Object::Stream(stream) => stream.set_content(decrypted),
+                Object::String(ref mut content, _) => *content = decrypted,
+                _ => {}
+            }
+        }
+
+        if let Ok(info_obj_id) = self.trailer.get(b"Info").and_then(Object::as_reference) {
+            if let Ok(info_dict) = self.get_object_mut(info_obj_id).and_then(Object::as_dict_mut) {
+                for (_, info_obj) in info_dict.iter_mut() {
+                    if let Ok(content) = encryption::decrypt_object(&key, info_obj_id, &*info_obj) {
+                        info_obj.as_str_mut().unwrap().clear();
+                        info_obj.as_str_mut().unwrap().extend(content);
+                    };
+                }
+            }
+        }
+
+        self.trailer.remove(b"Encrypt");
+        Ok(())
+    }
+
     /// Return the PDF document catalog, which is the root of the document's object graph.
     pub fn catalog(&self) -> Result<&Dictionary> {
         self.trailer
@@ -264,19 +349,34 @@ impl Document {
     pub fn get_page_contents(&self, page_id: ObjectId) -> Vec<ObjectId> {
         let mut streams = vec![];
         if let Ok(page) = self.get_dictionary(page_id) {
-            if let Ok(contents) = page.get(b"Contents") {
-                match *contents {
-                    Object::Reference(ref id) => {
-                        streams.push(*id);
-                    }
-                    Object::Array(ref arr) => {
-                        for content in arr {
-                            if let Ok(id) = content.as_reference() {
-                                streams.push(id)
+            let mut nb_deref = 0;
+            // Since we're looking for object ids, we can't use get_deref
+            // so manually walk any references in contents object
+            if let Ok(mut contents) = page.get(b"Contents") {
+                loop {
+                    match *contents {
+                        Object::Reference(id) => match self.objects.get(&id) {
+                            None | Some(Object::Stream(_)) => {
+                                streams.push(id);
+                            }
+                            Some(o) => {
+                                nb_deref += 1;
+                                if nb_deref < Self::DEREF_LIMIT {
+                                    contents = o;
+                                    continue;
+                                }
+                            }
+                        },
+                        Object::Array(ref arr) => {
+                            for content in arr {
+                                if let Ok(id) = content.as_reference() {
+                                    streams.push(id)
+                                }
                             }
                         }
+                        _ => {}
                     }
-                    _ => {}
+                    break;
                 }
             }
         }
@@ -352,15 +452,28 @@ impl Document {
         fn collect_fonts_from_resources<'a>(
             resources: &'a Dictionary, fonts: &mut BTreeMap<Vec<u8>, &'a Dictionary>, doc: &'a Document,
         ) {
-            if let Ok(font_dict) = resources.get(b"Font").and_then(Object::as_dict) {
-                for (name, value) in font_dict.iter() {
-                    let font = match *value {
-                        Object::Reference(id) => doc.get_dictionary(id).ok(),
-                        Object::Dictionary(ref dict) => Some(dict),
-                        _ => None,
-                    };
-                    if !fonts.contains_key(name) {
-                        font.map(|font| fonts.insert(name.clone(), font));
+            if let Ok(font) = resources.get(b"Font") {
+                let font_dict = match font {
+                    Object::Reference(ref id) => {
+                        doc.get_object(*id).and_then(Object::as_dict).ok()
+                    },
+                    Object::Dictionary(ref dict) => {
+                        Some(dict)
+                    },
+                    _ => {
+                        None
+                    }
+                };
+                if let Some(font_dict) = font_dict {
+                    for (name, value) in font_dict.iter() {
+                        let font = match *value {
+                            Object::Reference(id) => doc.get_dictionary(id).ok(),
+                            Object::Dictionary(ref dict) => Some(dict),
+                            _ => None,
+                        };
+                        if !fonts.contains_key(name) {
+                            font.map(|font| fonts.insert(name.clone(), font));
+                        }
                     }
                 }
             }
@@ -386,24 +499,91 @@ impl Document {
         let mut annotations = vec![];
         if let Ok(page) = self.get_dictionary(page_id) {
             match page.get(b"Annots") {
-                Ok(Object::Reference(ref id)) => self.get_object(*id)
+                Ok(Object::Reference(ref id)) => self
+                    .get_object(*id)
                     .and_then(Object::as_array)
                     .unwrap()
                     .iter()
                     .flat_map(Object::as_reference)
                     .flat_map(|id| self.get_dictionary(id))
                     .for_each(|a| annotations.push(a)),
-                Ok(Object::Array(ref a)) =>
-                    a.iter()
+                Ok(Object::Array(ref a)) => a
+                    .iter()
                     .flat_map(Object::as_reference)
                     .flat_map(|id| self.get_dictionary(id))
                     .for_each(|a| annotations.push(a)),
-                _ => {},
+                _ => {}
             }
         }
         annotations
     }
 
+    pub fn get_page_images(&self, page_id: ObjectId) -> Result<Vec<PdfImage>> {
+        let mut images = vec![];
+        if let Ok(page) = self.get_dictionary(page_id) {
+            let resources = self.get_dict_in_dict(page, b"Resources")?;
+            let xobject = self.get_dict_in_dict(resources, b"XObject")?;
+            for (_, xvalue) in xobject.iter() {
+                let id = xvalue.as_reference()?;
+                let xvalue = self.get_object(id)?;
+                let xvalue = xvalue.as_stream()?;
+                let dict = &xvalue.dict;
+                if dict.get(b"Subtype")?.as_name()? != b"Image" {
+                    continue;
+                }
+                let width = dict.get(b"Width")?.as_i64()?;
+                let height = dict.get(b"Height")?.as_i64()?;
+                let color_space = match dict.get(b"ColorSpace") {
+                    Ok(cs) => {
+                        match cs {
+                            Object::Array(array) => {
+                                Some(String::from_utf8_lossy(array[0].as_name()?).to_string())
+                            }
+                            Object::Name(name) => {
+                                Some(String::from_utf8_lossy(name).to_string())
+                            }
+                            _ => None
+                        }
+                    }
+                    Err(_) => None
+                };
+                let bits_per_component = match dict.get(b"BitsPerComponent") {
+                    Ok(bpc) => {
+                        Some(bpc.as_i64()?)
+                    }
+                    Err(_) => None
+                };
+                let mut filters = vec![];
+                if let Ok(filter) = dict.get(b"Filter") {
+                    match filter {
+                        Object::Array(array) => {
+                            for obj in array.iter() {
+                                let name = obj.as_name()?;
+                                filters.push(String::from_utf8_lossy(name).to_string());
+                            }
+                        }
+                        Object::Name(name) => {
+                            filters.push(String::from_utf8_lossy(name).to_string());
+                        }
+                        _ => {}
+                    }
+                };
+
+                images.push(PdfImage {
+                    id,
+                    width,
+                    height,
+                    color_space,
+                    bits_per_component,
+                    filters: Some(filters),
+                    content: &xvalue.content,
+                    origin_dict: &xvalue.dict,
+                });
+            }
+        }
+        Ok(images)
+    }
+
     pub fn decode_text(encoding: Option<&str>, bytes: &[u8]) -> String {
         if let Some(encoding) = encoding {
             info!("{}", encoding);
@@ -412,7 +592,7 @@ impl Document {
                 "MacRomanEncoding" => bytes_to_string(encodings::MAC_ROMAN_ENCODING, bytes),
                 "MacExpertEncoding" => bytes_to_string(encodings::MAC_EXPERT_ENCODING, bytes),
                 "WinAnsiEncoding" => bytes_to_string(encodings::WIN_ANSI_ENCODING, bytes),
-                "UniGB-UCS2-H" | "UniGB−UTF16−H" => UTF_16BE.decode(bytes, DecoderTrap::Ignore).unwrap(),
+                "UniGB-UCS2-H" | "UniGB−UTF16−H" => UTF_16BE.decode(bytes).0.to_string(),
                 "Identity-H" => "?Identity-H Unimplemented?".to_string(), // Unimplemented
                 _ => String::from_utf8_lossy(bytes).to_string(),
             }
@@ -428,7 +608,7 @@ impl Document {
                 "MacRomanEncoding" => string_to_bytes(encodings::MAC_ROMAN_ENCODING, text),
                 "MacExpertEncoding" => string_to_bytes(encodings::MAC_EXPERT_ENCODING, text),
                 "WinAnsiEncoding" => string_to_bytes(encodings::WIN_ANSI_ENCODING, text),
-                "UniGB-UCS2-H" | "UniGB−UTF16−H" => UTF_16BE.encode(text, EncoderTrap::Ignore).unwrap(),
+                "UniGB-UCS2-H" | "UniGB−UTF16−H" => UTF_16BE.encode(text).0.to_vec(),
                 "Identity-H" => vec![], // Unimplemented
                 _ => text.as_bytes().to_vec(),
             }
